/* label.c - label the objects in a thresholdable image
   ----------------------------------------------------

	Label a threholded image using 6 connectedness
		- See BKP Horn - Robot Vision p 65 - 89 




	labeling scheme

		+-+-+
		|D|C|
		+-+-+-+
		|B|A| |
		+-+-+-+
		  | | |
		  +-+-+
-------------------------------------------------------- */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "stats.h"


/* useful definintions */
#define		MIN( x, y )		(((x) < (y)) ? (x) : (y))

#define		UCHAR			unsigned char


#define		NO			0
#define		YES			-1
#define		NO_OBJECT		0	/* pixel value for no object present */



/* function prototypes */
static int IsObject( UCHAR *im, int nr, int nc, int r, int c, int threshold );




void LabelImage( UCHAR *im, UCHAR *lim, int nr, int nc, int threshold, int *nobj )
{
int r, c, i;
int LabelTable[20000];
int ntable, nobjects, newlabel;

	ntable = 0;
	LabelTable[0] = 0;

	for( r = 0; r < nr; r++ ) {
		for( c = 0; c < nc; c++ ) {
			if ( IsObject( im, nr, nc, r, c, threshold ) ) {	/* if A is an object */

				if ( IsObject( im, nr, nc, r, c-1, threshold ) ) {			/* if B is an object */
					lim[r * nc + c] = LabelTable[ lim[r * nc + c-1] ];		/*   tx lable from B to A */
					if ( IsObject( im, nr, nc, r-1, c, threshold ) ) {		/*   if C is obj too then */
						newlabel = MIN( LabelTable[ lim[(r-1) * nc + c] ],	/*      unify lables B and C */
								LabelTable[ lim[r * nc + (c-1)] ] );
						LabelTable[ lim[(r-1) * nc + c] ] = newlabel;
						LabelTable[ lim[r * nc + (c-1)] ] = newlabel;
						lim[r * nc + c] = newlabel;
					}
				}

				else if ( IsObject( im, nr, nc, r-1, c, threshold ) ) {		/* if C is an obj but B isn't */
					lim[r * nc + c] = LabelTable[ lim[(r-1) * nc + c] ];	/*    tx lable from C to A */
				}
				else {								/* found a new object */
					ntable++;
					lim[r * nc + c] = LabelTable[ ntable ] = ntable;	/*   label and put into table */
				}			

			} else {
				lim[r * nc + c] = NO_OBJECT;		/* A is not an object so leave it */
			}
		}
	}

	/* consolodate look-up table */
	for( i = 0; i <= ntable; i++ ) {
		newlabel = i;
		while ( newlabel != LabelTable[ newlabel ] )
			newlabel = LabelTable[ newlabel ];
		LabelTable[ i ] = newlabel;
	}

	/* run image through the look-up table */
	for( r = 0; r < nr; r++ )
		for( c = 0; c < nc; c++ )
			lim[r * nc + c] = LabelTable[ lim[r * nc + c] ];

	/* count up the objects in the image */
	for( i = 0; i <= ntable; i++ )
		LabelTable[i] = 0;

	for( r = 0; r < nr; r++ )
		for( c = 0; c < nc; c++ )
			LabelTable[ lim[r * nc + c] ]++;

	/* number the objects from 1 through n objects */
	*nobj = 0;
	LabelTable[0] = 0;
	for( i = 1; i <= ntable; i++ )
		if ( LabelTable[i] > 0 ) {
			(*nobj)++;
			LabelTable[i] = *nobj;
		}

	/* run through the look-up table again */
	for( r = 0; r < nr; r++ )
		for( c = 0; c < nc; c++ )
			lim[r * nc + c] = LabelTable[ lim[r * nc + c] ];
}

static int IsObject( UCHAR *im, int nr, int nc, int r, int c, int threshold )
{
	if ( r <  0  || c <  0  ) return NO;	/* out of bounds - too low */
	if ( r >= nr || c >= nc ) return NO;	/* out of bound - too high */
	if ( im[r * nc + c] < threshold ) return NO;	/* not an object pixel */
	return YES;
}



/*******************************************************************************/



void ImageStats( UCHAR *im, UCHAR *lim, int nr, int nc, int nobjects, struct StatSet *stattab )
{
int obj, pix, r, c;

	/* initialize params for counting */
	for( obj = 0; obj <= nobjects; obj++ ) {
		stattab[obj].npixels = 0;
		stattab[obj].pixmean = stattab[obj].pixstdv = 0.0;
		stattab[obj].maxrow = 0;
		stattab[obj].minrow = nr;
		stattab[obj].maxcol = 0;
		stattab[obj].mincol = nc;
		stattab[obj].rowmean = stattab[obj].rowstdv = 0.0;
		stattab[obj].colmean = stattab[obj].colstdv = 0.0;
	}


	for( r = 0; r < nr; r++ )
		for( c = 0; c < nc; c++ ) {
			obj = *lim++;
			pix = *im++;

			/* if ( obj == NO_OBJECT ) continue; */

			stattab[obj].npixels++;
			stattab[obj].pixmean += pix;
			stattab[obj].pixstdv += pix * pix;

			if ( r > stattab[obj].maxrow ) stattab[obj].maxrow = r;
			if ( r < stattab[obj].minrow ) stattab[obj].minrow = r;

			if ( c > stattab[obj].maxcol ) stattab[obj].maxcol = c;
			if ( c < stattab[obj].mincol ) stattab[obj].mincol = c;

			stattab[obj].rowmean += r;
			stattab[obj].rowstdv += r * r;
			stattab[obj].colmean += c;
			stattab[obj].colstdv += c * c;

		}

	for( obj = 0; obj <= nobjects; obj++ ) {
		stattab[obj].pixmean /= (float) stattab[obj].npixels;
		stattab[obj].pixstdv = sqrt( (stattab[obj].pixstdv / (float) stattab[obj].npixels) -
					(stattab[obj].pixmean * stattab[obj].pixmean) );

		stattab[obj].spanrow = stattab[obj].maxrow - stattab[obj].minrow;
		stattab[obj].spancol = stattab[obj].maxcol - stattab[obj].mincol;

		stattab[obj].rowmean /= (float) stattab[obj].npixels;
		stattab[obj].rowstdv = sqrt( (stattab[obj].rowstdv / (float) stattab[obj].npixels) -
					(stattab[obj].rowmean * stattab[obj].rowmean) );


		stattab[obj].colmean /= (float) stattab[obj].npixels;
		stattab[obj].colstdv = sqrt( (stattab[obj].colstdv / (float) stattab[obj].npixels) -
					(stattab[obj].colmean * stattab[obj].colmean) );
	}

}
